@@ -1,20 +1,17 @@
-// $Id: parser.c,v 1.18 2004/06/16 10:48:56 ensc Exp $    --*- c++ -*--
-
-// Copyright (C) 2002 Enrico Scholz <enrico.scholz@informatik.tu-chemnitz.de>
-//  
+// Copyright (C) 2002, 2003, 2004, 2008
+//               Enrico Scholz <enrico.scholz@informatik.tu-chemnitz.de>
+//
 // This program is free software; you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
-// the Free Software Foundation; version 2 of the License.
-//  
+// the Free Software Foundation; version 3 of the License.
+//
 // This program is distributed in the hope that it will be useful,
 // but WITHOUT ANY WARRANTY; without even the implied warranty of
 // MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 // GNU General Public License for more details.
-//  
+//
 // You should have received a copy of the GNU General Public License
-// along with this program; if not, write to the Free Software
-// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
-//  
+// along with this program. If not, see http://www.gnu.org/licenses/.
 
 #ifdef HAVE_CONFIG_H
 #  include <config.h>
@@ -29,6 +26,7 @@
 #include <string.h>
 #include <fcntl.h>
 #include <netinet/in.h>
+#include <sys/mman.h>
 
 #include "parser.h"
 #include "cfg.h"
@@ -36,6 +34,7 @@
 #include "wrappers.h"
 #include "compat.h"
 #include "output.h"
+#include "inet.h"
 
 #define tkEOF			(256)
 
@@ -56,13 +55,16 @@ static unsigned int	*CHARACTERS;
 #define chrIFNAME	0x400u
 #define chrFILENAME	0x800u
 #define chrBASEMOD	0x1000u
+#define chrVARNAME	0x2000u
 
 #define ensc_DHCP_FORWARDER_ULIMIT_H_I_KNOW_WHAT_I_DO
 #include "ulimit_codes.h"
 
+enum { MAX_VARNAME_SZ = 32 };
+
 inline static void
 initCharacterClassification(/*@out@*/unsigned int *chrs)
-    /*@requires maxSet(*chrs)==256@*/
+    /*@requires maxSet(*chrs)==257@*/
     /*@globals CHARACTERS@*/
     /*@modifies CHARACTERS, *chrs@*/
 {
@@ -75,17 +77,17 @@ initCharacterClassification(/*@out@*/unsigned int *chrs)
 
   /*@+charintliteral@*/
   for (c='0'; c<='9'; ++c) chrs[c] |= (chrDIGIT | chrNUMBER | chrIP |
-				       chrSYS | chrUSERNAME);
-  for (c='A'; c<='Z'; ++c) chrs[c] |= chrUPPERALPHA | chrSYS | chrUSERNAME;
-  for (c='a'; c<='z'; ++c) chrs[c] |= chrLOWALPHA   | chrSYS | chrUSERNAME;
+				       chrSYS | chrUSERNAME | chrVARNAME);
+  for (c='A'; c<='Z'; ++c) chrs[c] |= chrUPPERALPHA | chrSYS | chrUSERNAME | chrVARNAME;
+  for (c='a'; c<='z'; ++c) chrs[c] |= chrLOWALPHA   | chrSYS | chrUSERNAME | chrVARNAME;
   chrs['\r'] |= chrNL;
   chrs['\n'] |= chrNL;
-  
+
   chrs['\t'] |= chrBLANK;
   chrs[' ']  |= chrBLANK;
-  chrs['.']  |= chrSYS | chrUSERNAME | chrIP;
-  chrs['_']  |= chrSYS | chrUSERNAME;
-  chrs['-']  |= chrSYS | chrUSERNAME;
+  chrs['.']  |= chrSYS | chrUSERNAME | chrIP | chrVARNAME;
+  chrs['_']  |= chrSYS | chrUSERNAME | chrVARNAME;
+  chrs['-']  |= chrSYS | chrUSERNAME | chrVARNAME;
   chrs[':']  |= chrSYS;
   chrs['/']  |= chrFILENAME;
 
@@ -108,9 +110,10 @@ isCharType(/*@sef@*/int c, /*@sef@*/unsigned int type) /*@*/
 }
 
 /*@unchecked@*/static int		look_ahead = tkEOF;
-/*@checkmod@*/static int		fd;
 /*@unchecked@*/static unsigned int	line_nr, col_nr;
 /*@null@*/static char const		*filename = 0;
+char const				*cfg_ptr;
+char const				*cfg_end;
 
   /*@noreturn@*//*@unused@*/
 static void scEXITFATAL(/*@observer@*//*@sef@*/char const *msg);
@@ -123,7 +126,7 @@ exitFatal(/*@observer@*/char const msg[],
 	  register size_t len) __attribute__ ((noreturn)) /*@*/ ;
 
 inline static void
-exitFatal(char const msg[], register size_t len) 
+exitFatal(char const msg[], register size_t len)
 {
     /*@-internalglobs@*//*@-globs@*/
   if (filename!=0) (void)write(2, filename, strlen(filename));
@@ -146,21 +149,20 @@ setNext()
     /*@modifies col_nr, line_nr, look_ahead@*/
 {
   char			c;
-  int			cnt;
 
   ++col_nr;
-  cnt = TEMP_FAILURE_RETRY(read(fd, &c, 1));
-  if (cnt==-1) scEXITFATAL("read() failed");
+  if (cfg_ptr == cfg_end)
+    look_ahead = tkEOF;
+  else {
+    c = *cfg_ptr;
+    ++cfg_ptr;
 
-  if (cnt>0) {
     switch (c) {
       case '\n'	:  ++line_nr; col_nr = 0; break;
       case '\t'	:  col_nr = (col_nr+7)/8 * 8;
     }
     look_ahead = static_cast(int)(c);
   }
-  else if (cnt==0) look_ahead = tkEOF;
-  else { assert(false); }
 }
 
 inline static int
@@ -179,7 +181,7 @@ match(char /*@alt int@*/ c)
     /*@modifies look_ahead@*/
 {
   register int		got = getLookAhead();
-  
+
   if (got==tkEOF) scEXITFATAL("unexpected EOF while parsing");
   if (got!=c)     scEXITFATAL("unexpected symbol");
 
@@ -201,19 +203,22 @@ newInterface(struct InterfaceInfoList *ifs)
 {
   size_t		new_len;
   struct InterfaceInfo	*result;
-  
+
   ++ifs->len;
-  
+
   new_len  = ifs->len * (sizeof(*ifs->dta));
   ifs->dta = static_cast(struct InterfaceInfo*)(Erealloc(ifs->dta, new_len));
 
   assert(ifs->dta!=0);
-  
-  result            = &ifs->dta[ifs->len - 1];
+
+  result              = &ifs->dta[ifs->len - 1];
   memset(result, 0, sizeof result);
-  result->if_ip     = INADDR_NONE;
-  result->sender_fd = -1;
-  
+
+  result->if_ip       = INADDR_NONE;
+  result->port_client = htons(DHCP_PORT_CLIENT);
+  result->port_server = htons(DHCP_PORT_SERVER);
+  result->sender_fd   = -1;
+
   return result;
 }
 
@@ -223,7 +228,7 @@ newServer(struct ServerInfoList *servers)
 {
   size_t		new_len;
   struct ServerInfo	*result;
-  
+
   ++servers->len;
   new_len      = servers->len * (sizeof(*servers->dta));
   servers->dta = static_cast(struct ServerInfo *)(Erealloc(servers->dta, new_len));
@@ -232,7 +237,7 @@ newServer(struct ServerInfoList *servers)
   result	 = &servers->dta[servers->len - 1];
   result->type   = svUNICAST;
   memset(&result->info, 0, sizeof result->info);
-  
+
   return result;
 }
 
@@ -303,23 +308,23 @@ matchEOL()
     /*@modifies look_ahead@*/
 {
   register int		state = 0xFF00;
-  
+
   while (state!=0xFFFF) {
     int			c = getLookAhead();
-    
+
     switch (state) {
       case 0xFF00	:
 	if      (isCharType(c, chrBLANK)) {}
 	else if (isCharType(c, chrNL))    { state = 0; }
 	else goto err;
-	
+
 	match(c);
 	break;
-	
+
       case 0		:
 	if (isCharType(c, chrNL)) match(c);
 	else                      state = 0xFFFF;
-	
+
 	break;
 
       default		:  assert(false); goto err;
@@ -381,12 +386,50 @@ readString(/*@out@*/char *buffer, size_t len, unsigned int char_class)
 }
 /*@=charintliteral@*/
 
+inline static size_t
+readStringExpanded(/*@out@*/char *buffer, size_t len, unsigned int char_class)
+{
+  int		c = getLookAhead();
+  if (c == '\\')
+    match(c);
+  else if (c == '$') {
+    size_t	l;
+    size_t	i;
+    char	var[MAX_VARNAME_SZ+1];
+    char const	*tmp;
+    match(c);
+    l = readString(var, sizeof var, chrVARNAME);
+
+    if (l==0)
+      scEXITFATAL("empty variable name");
+
+    tmp = getenv(var);
+
+    if (!tmp)
+      scEXITFATAL("failed not expand variable");
+
+    l = strlen(tmp);
+    if (l >= len)
+      scEXITFATAL("expanded value too long");
+
+    memcpy(buffer, tmp, l+1);
+    for (i = 0; i < l; ++i) {
+      if (!isCharType(buffer[i], char_class))
+	scEXITFATAL("bad chars in expanded variable");
+    }
+
+    return l;
+  }
+
+  return readString(buffer, len, char_class);
+}
+
 inline static void
 readFileName(/*@out@*/char buffer[], size_t len)
     /*@globals fd, look_ahead@*/
     /*@modifies look_ahead, *buffer@*/
 {
-  if (readString(buffer, len, chrFILENAME)==0)
+  if (readStringExpanded(buffer, len, chrFILENAME)==0)
     scEXITFATAL("Invalid filename");
 }
 
@@ -406,7 +449,7 @@ readLimit()
 {
   char		buffer[128];
   size_t	i;
-  
+
   (void)readString(buffer, sizeof buffer, chrALPHA);
   for (i=0; i<sizeof(ULIMIT_CODES)/sizeof(ULIMIT_CODES[0]); ++i) {
     if (strcmp(ULIMIT_CODES[i].name, buffer)==0) return ULIMIT_CODES[i].code;
@@ -430,12 +473,12 @@ readLong()
   while (state!=0xFFFF && ptr+2<&value[sizeof(value)]) {
     /*@=ptrarith@*/
     int		c = getLookAhead();
-    
+
     switch (state) {
       case 0		:
       {
 	bool		do_match = true;
-	
+
 	switch (c) {
 	  case '-'	:  break;
 	  case '0'	:  state = 0x0200; break;
@@ -481,7 +524,7 @@ readLong()
 
   assert(ptr<&value[sizeof(value)]);
   assertDefined(value);
-  
+
   *ptr   = '\0';
   result = strtol(value, &ptr, 0);
   if (*ptr!='\0') goto err;
@@ -500,6 +543,38 @@ readInteger()
   return static_cast(int)(readLong());
 }
 
+inline static int
+readIntegerExpanded()
+{
+  char			buffer[1024];
+  char			*err_ptr;
+  long			res;
+
+  (void)readStringExpanded(buffer, sizeof buffer, chrNUMBER);
+
+  res = strtol(buffer, &err_ptr, 10);
+  if (res == LONG_MIN || res == LONG_MAX)
+    scEXITFATAL("conversion under/overflow");
+
+  if (err_ptr == buffer || *err_ptr)
+    scEXITFATAL("failed to convert number");
+
+  return res;
+}
+
+inline static uint16_t
+readPort(uint16_t dflt)
+{
+  int			res = readIntegerExpanded();
+
+  if (res < 0 || res >= (1<<16))
+    scEXITFATAL("port number out of range");
+  else if (res == 0)
+    res = dflt;
+
+  return res;
+}
+
 inline static rlim_t
 readLimitVal()
     /*@globals fd, look_ahead@*/
@@ -529,7 +604,7 @@ readIfname(/*@out@*/char *iface)
     /*@globals fd, look_ahead@*/
     /*@modifies look_ahead, *iface@*/
 {
-  if (readString(iface, IFNAMSIZ, chrIFNAME)==0)
+  if (readStringExpanded(iface, IFNAMSIZ, chrIFNAME)==0)
     scEXITFATAL("Invalid interface name");
 }
 
@@ -540,7 +615,7 @@ readIp(/*@out@*/struct in_addr	*ip)
     /*@modifies look_ahead, *ip@*/
 {
   char			buffer[1024];
-  (void)readString(buffer, sizeof buffer, chrIP);
+  (void)readStringExpanded(buffer, sizeof buffer, chrIP);
 
   assertDefined(buffer);	// ptr is an alias for buffer
   if (inet_aton(buffer, ip)==0) scEXITFATAL("Invalid IP");
@@ -556,7 +631,7 @@ readBool(/*@out@*/bool *val)
 
   while (state!=0xFFFF) {
     int		c = getLookAhead();
-    
+
     switch (state) {
       case 0	:
 	switch (c) {
@@ -622,25 +697,44 @@ parse(/*@in@*/char const		fname[],
   struct in_addr	ip;
   bool			has_clients;
   bool			has_servers;
+  uint16_t		port_client;
+  uint16_t		port_server;
   bool			allow_bcast;
   struct {
       int		code;
       rlim_t		val;
-  }			ulimit;
-  
+  }			ulimit = { 0,0 };
+  int			fd;
+  struct stat		st;
+  char			*cfg_start;
+
 
   filename = fname;
   line_nr  = 1u;
   col_nr   = 1u;
-  
+
   fd = open(filename, O_RDONLY);
   if (fd==-1) {
     perror("open()");
     exit(1);
   }
 
+  if (fstat(fd, &st)<0) {
+    perror("fstat()");
+    exit(1);
+  }
+
+  cfg_start = mmap(NULL, st.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
+  if (!cfg_start) {
+    perror("mmap()");
+    exit(1);
+  }
+
+  cfg_ptr = cfg_start;
+  cfg_end = cfg_start + st.st_size;
+
   initCharacterClassification(chrs);
-  
+
   while (state!=0xFFFF) {
     int		c = getLookAhead();
 
@@ -653,7 +747,7 @@ parse(/*@in@*/char const		fname[],
 	break;
 
       case 0xFFFE	:  matchEOL(); state=0; break;
-	  
+
       case 0		:
 	if      (isCharType(c, chrBLANK)) match(c);
 	else if (isCharType(c, chrNL))    matchEOL();
@@ -666,13 +760,47 @@ parse(/*@in@*/char const		fname[],
 	  case 'g'	:  state = 0x0500; break;
 	  case 'c'	:  state = 0x0600; break;
 	  case 'l'	:  state = 0x0700; break;
-	  case 'p'	:  state = 0x0800; break;
+	  case 'p'	:  state = 0x0890; break;
 	  case tkEOF	:  state = 0xFFFF; break;
 	  default	:  goto err;
 	}
 	if (state!=0 && state!=0xFFFF) match(c);
 	break;
 
+      case 0x0890:
+	switch (c) {
+	  case 'i':  state = 0x0800; break; /* pidfile */
+	  default:
+	  case 'o':  state = 0x0850; break; /* ports */
+	}
+	break;
+
+      case 0x0850:
+	matchStr("orts");   readBlanks();
+	readIfname(ifname); readBlanks();
+	port_client = readPort(DHCP_PORT_CLIENT); readBlanks();
+	port_server = readPort(DHCP_PORT_SERVER);
+	++state;
+	break;
+
+      case 0x0851: {
+	struct InterfaceInfo	*iface;
+
+	// Reachable from state 0x0850 only
+	assertDefined(ifname);
+	assertDefined(port_client);
+	assertDefined(port_server);
+
+	iface = searchInterface(&cfg->interfaces, ifname);
+	if (port_client)
+	  iface->port_client = htons(port_client);
+	if (port_server)
+	  iface->port_server = htons(port_server);
+
+	state = 0xFFFE;
+	break;
+      }
+
       case 0x0800	:
 	matchStr("idfile"); readBlanks();
 	++state;
@@ -682,7 +810,7 @@ parse(/*@in@*/char const		fname[],
 	readFileName(cfg->pidfile_name, sizeof(cfg->pidfile_name));
 	state = 0xFFFE;
 	break;
-	
+
       case 0x0700	:
 	matchStr("og");
 	++state;
@@ -706,7 +834,7 @@ parse(/*@in@*/char const		fname[],
 	cfg->loglevel = readInteger();
 	state = 0xFFFE;
 	break;
-	
+
       case 0x0400	:
 	switch (c) {
 	  case 's'	:  state = 0x1401; break;
@@ -751,7 +879,7 @@ parse(/*@in@*/char const		fname[],
 
 	  // can be reached only from state 0x0501 where 'name' was read
 	assertDefined(name);
-	
+
 	nr = strtol(name, &err_ptr, 0);
 	if (*err_ptr!='\0') state += 0x11;
 	else                state += 0x01;
@@ -762,7 +890,7 @@ parse(/*@in@*/char const		fname[],
       case 0x0403	:
 	  // Can be reached from state 0x0402 only
 	assertDefined(&nr);
-	
+
 	cfg->uid = nr;
 	state = 0xFFFE;
 	break;
@@ -770,7 +898,7 @@ parse(/*@in@*/char const		fname[],
       case 0x0413	:
 	  // Can be reached from state 0x0402 only
 	assertDefined(name);
-	
+
 	cfg->uid = Egetpwnam(name)->pw_uid;
 	state = 0xFFFE;
 	break;
@@ -778,7 +906,7 @@ parse(/*@in@*/char const		fname[],
       case 0x0503	:
 	  // Can be reached from state 0x0502 only
 	assertDefined(&nr);
-	
+
 	cfg->gid = nr;
 	state = 0xFFFE;
 	break;
@@ -786,7 +914,7 @@ parse(/*@in@*/char const		fname[],
       case 0x0513	:
 	  // Can be reached from state 0x0502 only
 	assertDefined(name);
-	
+
 	cfg->gid = Egetgrnam(name)->gr_gid;
 	state = 0xFFFE;
 	break;
@@ -821,7 +949,7 @@ parse(/*@in@*/char const		fname[],
 	readBool(&allow_bcast);
 	state = 0x110;
 	break;
-	
+
       case 0x110	:
       {
 	struct InterfaceInfo *	iface = newInterface(&cfg->interfaces);
@@ -831,8 +959,8 @@ parse(/*@in@*/char const		fname[],
 	assertDefined(&has_clients);
 	assertDefined(&has_servers);
 	assertDefined(&allow_bcast);
-	
-	
+
+
 	strcpy(iface->name, ifname);
 	iface->aid[0]      = '\0';
 	iface->has_clients = has_clients;
@@ -859,19 +987,19 @@ parse(/*@in@*/char const		fname[],
 	  // Reachable from state 0x0120 only
 	assertDefined(ifname);
 	assertDefined(&ip);
-	iface = searchInterface(&cfg->interfaces, ifname);	
+	iface = searchInterface(&cfg->interfaces, ifname);
 
 	iface->if_ip = ip.s_addr;
 	state = 0xFFFE;
 	break;
       }
 
-	
+
       case 0x200	:
 	matchStr("ame");      readBlanks();
 	state = 0x201;
 	break;
-	
+
       case 0x201	:
 	readIfname(ifname);   readBlanks();
 	readIfname(agent_id);
@@ -924,13 +1052,15 @@ parse(/*@in@*/char const		fname[],
 	struct ServerInfo	*server = newServer(&cfg->servers);
 	struct InterfaceInfo	*iface  = 0;
 
+	  // Reachable from state 0x311 only
+	assertDefined(ifname);
 	if (ifname[0]!='\0')
 	  iface = searchInterface(&cfg->interfaces, ifname);
 
 	server->type            = svUNICAST;
 	server->iface           = iface;
 	server->info.unicast.ip = ip;
-	
+
 	state = 0xFFFE;
 	break;
       }
@@ -954,7 +1084,7 @@ parse(/*@in@*/char const		fname[],
 	iface->need_mac    = true;
 	server->type       = svBCAST;
 	server->iface      = iface;
-	
+
 	state = 0xFFFE;
 	break;
       }
@@ -963,6 +1093,7 @@ parse(/*@in@*/char const		fname[],
     }
   }
 
+  munmap(cfg_start, st.st_size);
   Eclose(fd);
   return;
 
